<?php

namespace App\Services;

use Carbon\Carbon;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Session;

class BookingSession
{
    public string $id;
    public Carbon $start;
    public $products;
    public int $clubs;
    public $tableReservation;
    public $foodDelivery;
    public array $guests;
    public $thawani_session_id;
    public $golf_manager_session;

    public function __construct($data = [])
    {
        $this->id = $data['id'] ?? null;
        $this->start = isset($data['start']) ? Carbon::create($data['start']) : Carbon::now();
        $this->tableReservation = isset($data['tableReservation']) ? $data['tableReservation'] : null;
        $this->foodDelivery = isset($data['foodDelivery']) ? $data['foodDelivery'] : null;
        $this->products = $data['products'] ?? [];
        $this->clubs = $data['clubs'] ?? 0;
        $this->golf_manager_session = $data['golf_manager_session'] ?? null;
        $this->thawani_session_id = $data['thawani_session_id'] ?? null;

        $this->guests = $data['guests'];
    }

    public function __get($name)
    {
        switch ($name) {
            case "salelines":
                return $this->getSaleLinesAttribute();
        }

        return null;
    }

    /**
     * Get function to return instance of self
     *
     * @return self | null
     */
    public static function get()
    {
        $booking = session('booking');

        if (!$booking) {
            return null;
        }

        return new self($booking);
    }

    public static function create($data = []): self
    {
        $data['guests'] = array_merge([
            'adults' => [],
            'juniors' => [],
        ], $data['guests'] ?? [
            'adults' => [],
            'juniors' => [],
        ]);

        if (!Auth::getUser()->isGuest()) {
            $prices = GolfManagerService::getPrices(session('user_meta')['idTags']);

            $data['guests']['adults'][] = [
                'id' => Auth::getUser()->id,
                'type' => 'member',
                'price' => $prices['adults'],
                'first_name' => Auth::getUser()->name,
                'self' => true,
            ];
        } else {
            $guest = GolfManagerService::searchUsers('Guest User')->first();

            if ($guest) {
                $prices = GolfManagerService::getPrices($guest['idTags'], true);
            }

            $data['guests']['adults'][] = [
                'type' => 'guest',
                'price' => $prices['adults'] ?? 0,
                'self' => true,
            ];
        }

        session([
            'booking' => $data,
        ]);

        return self::get();
    }

    public function update($data = []): self
    {
        session([
            'booking' => array_merge(session('booking', []), $data),
        ]);

        return self::get();
    }

    public static function delete(): void
    {
        Session::forget('booking');
    }

    public function hasGuests(): bool
    {
        if (!session('booking.guests.adults', 0) && !session('booking.guests.adults', 0)) {
            return false;
        }

        return true;
    }

    public function getPrimaryGuest()
    {
        return $this->guests['adults'][0] ?? null;
    }

    public function mailingList(): array
    {
        $mailingList = [];

        foreach ($this->guests['adults'] as $guest) {
            $mailingList[] = $guest['email'];
        }

        return $mailingList;
    }

    public function getReservationIds()
    {
        if (!$this->golf_manager_session) {
            return [];
        }

        $reservations = collect($this->golf_manager_session['reservations']);

        return $reservations->map(function ($reservation) {
            return $reservation['id'];
        });
    }

    public function getSaleLinesAttribute(): array
    {
        $products = [];

        $adults = count($this->guests['adults']) ?? 0;
        $juniors = count($this->guests['juniors']) ?? 0;
        $clubs = $this->clubs ?? 0;

        if ($adults) {
            $products[] = [
                'idProduct' => 99,
                'quantity' => $adults,
            ];
        }

        if ($juniors) {
            $products[] = [
                'idProduct' => 95,
                'quantity' => $juniors,
            ];
        }

        if ($clubs) {
            $products[] = [
                'idProduct' => 13,
                'quantity' => $clubs,
            ];
        }

        foreach (['food', 'beverages'] as $itemType) {
            if (count($this->basket[$itemType] ?? [])) {
                foreach ($this->basket[$itemType] as $key => $item) {
                    $itemProduct = [
                        'idProduct' => null,
                        'quantity' => 0,
                    ];

                    foreach ($this->basket[$itemType][$key]['orders'] as $order) {
                        $itemProduct['idProduct'] = $order['id'];
                        $itemProduct['quantity'] += $order['quantity'];
                    }

                    $products[] = $itemProduct;
                }
            }
        }

        return $products;
    }

    /**
     * This converts the current booking to an array that can be passed to Thawani
     *
     * $options['foodOnly'] Filter for only food products
     */
    public function getThawaniProducts($options = []): array
    {
        $products = [];

        $bookingProducts = GolfManagerService::getBookingProducts();

        $clubProduct = $bookingProducts->first(function ($bookingProduct) {
            return $bookingProduct['name'] === '18 Hole Rental clubs';
        });

        $clubs = $this->products['clubs'] ?? 0;

        if (!isset($options['foodOnly']) || !$options['foodOnly']) {
            foreach ($this->guests['adults'] as $adult) {
                if ($adult['price'] > 0) {
                    $products[] = [
                        'name' => 'Adult Guest Golfing Session',
                        'quantity' => 1,
                        'unit_amount' => $adult['price'],
                    ];
                }
            }

            foreach ($this->guests['juniors'] as $junior) {
                if ($junior['price'] > 0) {
                    $products[] = [
                        'name' => 'Junior Guest Golfing Session',
                        'quantity' => 1,
                        'unit_amount' => $junior['price'],
                    ];
                }
            }

            if ($clubs) {
                if ($clubProduct['price'] > 0) {
                    $products[] = [
                        'name' => 'Hire Clubs',
                        'quantity' => $clubs,
                        'unit_amount' => $clubProduct ? $clubProduct['price'] * 1000 : 15000,
                    ];
                }
            }
        }

        foreach (['food', 'beverages'] as $itemType) {
            if (count($this->products[$itemType] ?? [])) {
                foreach ($this->products[$itemType] as $key => $item) {
                    $itemProduct = [
                        'name' => $item['name'],
                        'quantity' => $item['quantity'],
                        'unit_amount' => floor($item['price'] * 1000),
                    ];

                    $products[] = $itemProduct;
                }
            }
        }

        return $products;
    }
}
