<?php

namespace App\Services;

use Carbon\Carbon;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Cache;
use Illuminate\Support\Facades\Crypt;
use Illuminate\Support\Facades\Http;
use Illuminate\Support\Str;

class GolfManagerService
{
    public static function getCredentials($isAdmin)
    {
        return [
            'username' => $isAdmin
                ? config('services.golfmanager.admin.username')
                : (Auth::user() && !Auth::getUser()->isGuest()
                    ? Auth::user()->email
                    : config('services.golfmanager.consumer.username')),
            'password' => $isAdmin
                ? config('services.golfmanager.admin.password')
                : (Auth::user() && !Auth::getUser()->isGuest()
                    ? Crypt::decrypt(Auth::user()->password)
                    : config('services.golfmanager.consumer.password')),
            'url' => $isAdmin || (Auth::user() && Auth::getUser()->isGuest()) ? config('services.golfmanager.url') : config('services.golfmanager.consumer_url'),
        ];
    }

    public static function get($url, $query = [], $isAdmin = false)
    {
        $query = array_merge($query, [
            'tenant' => 'almouj',
        ]);

        $credentials = self::getCredentials($isAdmin);

        return Http::withBasicAuth(
            $credentials['username'],
            $credentials['password'],
        )->get($credentials['url'] . $url, $query);
    }

    public static function post($url, $data = [], $isAdmin = false)
    {
        $query = urldecode(http_build_query(['tenant' => 'almouj']));

        $credentials = self::getCredentials($isAdmin);

        $client = new \GuzzleHttp\Client(['http_errors' => false]);
        $response = $client->request('POST', $credentials['url'] . $url . '?' . $query, [
            'headers' => [
                'Content-Type: application/json',
            ],
            'auth' => [
                $credentials['username'],
                $credentials['password'],
            ],
            'form_params' => $data,
        ]);

        return json_decode(preg_replace('/[\x00-\x1F\x80-\xFF]/', '', $response->getBody()), true);
    }

    public static function findUser($email)
    {
        return self::get('/clientsFull', [
            'email' => $email,
        ], true)->json()[0] ?? null;
    }

    public static function getUsers(): Collection
    {
        $response = self::get('/clientsFull', [], true);

        if ($response->status() != 200) {
            return collect([]);
        }

        return collect($response->json());
    }

    public static function getUserById($id): array
    {
        $response = self::get('/clientsFull', [
            'id' => $id,
        ], true);

        if ($response->status() != 200) {
            return null;
        }

        return collect($response->json())->first();
    }

    public static function searchUsers($search): Collection
    {
        $response = self::get('/clientsFull', [
            'search' => $search,
        ], true);

        if ($response->status() != 200) {
            return collect([]);
        }

        return collect($response->json());
    }

    public static function login($email, $password)
    {
        $response = Http::withBasicAuth(
            $email,
            $password,
        )->get(config('services.golfmanager.consumer_url') . '/tenant');

        if ($response->status() == 200) {
            return self::findUser($email);
        }

        return null;
    }

    public static function resetPassword($id, $password)
    {
        return self::post('/saveClient', [
            'data' => json_encode([
                'id' => $id,
                'password' => $password,
            ]),
        ], true);
    }

    public static function getAvailability($data)
    {
        $start = Carbon::create($data['start']) ?? Carbon::tomorrow()->setHours(12)->setMinutes(0)->setSeconds(0);

        $response = self::get('/searchAvailability', [
            'start' => $start->format('Y-m-d') . 'T' . $start->format('H:i'),
            'slots' => 1,
        ]);

        if ($response->status() != 200) {
            return [];
        }

        return $response->json();
    }

    /**
     * @param $product array [id, name]
     */
    public static function getProductPrice($product)
    {
        $start = session('search_availability', null) ? Carbon::parse(session('search_availability')) : now();

        $weektype = ($start->dayOfWeek == 5 || $start->dayOfWeek == 6) ? 'weekend' : 'weekday';
        $season = $start->month >= 6 && $start->month <= 9 ? 'summer' : 'winter';

        Cache::flush();
        $prices = Cache::remember('prices-' . $product['id'], 6000, function () use ($product) {
            return collect(self::get('/prices', [
                'idProduct' => $product['id'],
            ], true)->json());
        });

        $price = $prices->first(function ($price) use ($weektype, $season, $start) {
            return
                Str::contains(Str::lower($price['name']), $weektype)
                && Str::contains(Str::lower($price['name']), $season)
                && $start->isAfter(Carbon::parse($price['start']))
                && $start->isBefore(Carbon::parse($price['end']));
        });

        return ($price ? $price['value'] : 0) * 1000;
    }

    public static function getBookingTypes($userTags = '')
    {
        $response = self::get('/bookingsTypes', [
            'includeTags' => true,
        ], true);

        if ($response->status() != 200) {
            return collect([]);
        }

        Cache::flush();

        $bookingTypes = Cache::remember('booking_types' . $userTags, 6000, function () use ($response) {
            return collect($response->json());
        });

        return $bookingTypes;
    }

    private static function isGuestType($type)
    {
        return Str::contains(Str::lower($type['name']), 'visitor') || Str::contains(Str::lower($type['name']), 'guest');
    }

    private static function isMemberType($type)
    {
        return Str::contains(Str::lower($type['name']), 'gf')
            && Str::contains(Str::lower($type['name']), '18h');
    }

    private static function isValidType($type)
    {
        return Str::contains(Str::lower($type['name']), 'gf')
            && Str::contains(Str::lower($type['name']), '18h');
    }

    private static function isJuniorType($type)
    {
        return Str::contains(Str::lower($type['name']), 'junior');
    }

    public static function getUserBookingTypes($userTags = null, $isGuest = false)
    {
        $userTags = explode(',', ($userTags ? $userTags : session('user_meta')['idTags']));

        return GolfManagerService::getBookingTypes(implode(',', $userTags))->filter(function ($type) use ($userTags, $isGuest) {
            if (($isGuest && !self::isGuestType($type)) || (!$isGuest && self::isGuestType($type))) {
                return;
            }

            if (!self::isValidType($type)) {
                return;
            }

            if ($type['idResourceType'] != 1) {
                return false;
            }

            if (!array_intersect(explode(',', $type['idTags'] ?? ''), $userTags)) {
                return false;
            }

            if (self::isJuniorType($type)) {
                return false;
            }

            return true;
        });
    }

    /**
     * @param $userTags string eg:1,2,3,4
     */
    public static function getPrices($userTags = null, $forGuest = false)
    {
        $prices = [];
        $userTags = explode(',', ($userTags ? $userTags : session('user_meta')['idTags']));

        GolfManagerService::getBookingTypes(implode('', $userTags))->each(function ($type) use (&$prices, $userTags, $forGuest) {
            if (($forGuest && !self::isGuestType($type)) || (!$forGuest && self::isGuestType($type))) {
                return;
            }

            if (!self::isValidType($type)) {
                return;
            }

            if (!array_intersect(explode(',', $type['idTags'] ?? ''), $userTags)) {
                return;
            }

            if (Str::contains(Str::lower($type['name']), 'junior')) {
                $prices['juniors'] = self::getProductPrice(['id' => $type['idProduct']]);
            } else {
                $prices['adults'] = self::getProductPrice(['id' => $type['idProduct']]);
            }
        });

        if (!isset($prices['juniors'])) {
            $prices['juniors'] = $prices['adults'] ?? 0;
        }

        return $prices;
    }

    public static function getProducts()
    {
        $response = self::get('/products', [
            'sellOnline' => 'true',
        ], true);

        if ($response->status() != 200) {
            return collect([]);
        }

        return collect($response->json());
    }

    public static function getBookingProducts()
    {
        $response = self::get('/products', [
            'search' => '18',
        ], true);

        if ($response->status() != 200) {
            return collect([]);
        }

        return Cache::remember('booking-products', 6000, function () use ($response) {
            return collect($response->json());
        });
    }

    public static function getReservations($query)
    {
        return collect(self::get('/reservations', [
            'start' => $query['start']->setTimezone('UTC')->toIso8601String(),
            'end' => $query['end']->setTimezone('UTC')->toIso8601String(),
        ], true)->json())->filter(function ($reservation) {
            if (isset($reservation['subfamilyName']) && $reservation['subfamilyName'] == 'CLUB HIRE') {
                return false;
            }

            return true;
        });
    }

    public static function getReservationTypes()
    {
        $response = self::get('/reservationtypes', [], true);

        if ($response->status() != 200) {
            return collect([]);
        }

        return Cache::remember('reservation_types', 6000, function () use ($response) {
            return collect($response->json());
        });
    }

    public static function makeReservation(BookingSession $booking)
    {
        $guest = null;

        if (Auth::getUser()->isGuest()) {
            $guest = GolfManagerService::searchUsers('Guest User')->first();
        }

        $type = GolfManagerService::getUserBookingTypes($guest ? $guest['idTags'] : session('user_meta')['idTags'], Auth::getUser()->isGuest())->first();

        $reservation = [
            'idType' => $type['id'],
            'start' => $booking->start->toIso8601String(),
            'end' => $booking->start->copy()->addHour()->toIso8601String(),
            'timeout' => now()->addMinutes(10),
        ];

        if (Auth::getUser()->isGuest()) {
            if ($primaryGuest = $booking->getPrimaryGuest()) {
                $reservation['name'] = $primaryGuest['first_name'] . ' ' . $primaryGuest['last_name'];
                $reservation['email'] = $primaryGuest['email'];
            }
        } else {
            $reservation['name'] = Auth::user()->golf_manager_id;
            $reservation['email'] = Auth::user()->email;
        }

        return self::post('/makeReservation', [
            'reservations' => json_encode([
                $reservation,
            ]),
        ]);
    }

    public static function confirmReservation(BookingSession $booking)
    {
        $reservations = collect($booking->golf_manager_session['reservations'] ?? []);

        $ids = $reservations->map(function ($reservation) {
            return $reservation['id'];
        });

        return self::post('/confirmReservation', [
            'ids' => json_encode($ids),
            'paymentMethod' => 'Thawani',
        ]);
    }
}
