import React from 'react';
import { Settings, DateTime } from 'luxon';
import { useEffect, useState } from 'react';
import Main from '../../layouts/Main';
import { Inertia } from '@inertiajs/inertia';
import { v4 } from 'uuid';
import Alpha from '../../components/typography/Alpha';
import axios from 'axios';
import { CustomSelect } from '../../components/customSelect';
import NProgress from 'nprogress'

const Index = ({ user, booking, search }) => {
  const [teeDate, setTeeDate] = useState(
    search ? DateTime.fromISO(search).toFormat('yyyy-MM-dd') : getNextDay()
  );
  const [teeTime, setTeeTime] = useState(
    search
      ? { label: DateTime.fromISO(search).toFormat('ha'), value: DateTime.fromISO(search).toFormat('HH:mm') }
      : { label: getNextTimeAsFormat('ha'), value: getNextTimeAsFormat('HH:mm') }
  );

  const [slots, setSlots] = useState([]);
  const [loading, setLoading] = useState(false);
  const [lastUpdated, setLastUpdated] = useState(DateTime.now());

  useEffect(async () => {
    setLoading(true);
    NProgress.start()
    let results = (await axios('/api/availability', {
      params: {
        client: user.golf_manager_id,
        start: DateTime.fromFormat(`${teeDate} ${teeTime.value}`, 'yyyy-MM-dd HH:mm').toISO(),
      },
    }))?.data;

    setSlots(results?.slots);

    NProgress.done()
    setLoading(false);
  }, [lastUpdated]);

  function getNextDay() {
    let now = DateTime.now();
    return now.hour >= 16 ? now.plus({ days: 1 }).toFormat('yyyy-MM-dd') : now.toFormat('yyyy-MM-dd')
  }

  function getNextTimeAsFormat(format) {
    let now = DateTime.now();

    if (now.hour >= 16) {
      now = now.set({ hour: 7, minute: 0, second: 0 })
    }

    return now.plus({ days: 1 }).toFormat(format).toLowerCase()
  }

  function reserveSlot (start) {
    Inertia.post('/booking', {
      id: v4(),
      start: start,
    });
  }

  const timeOptions = () => {
    let time = DateTime.now().set({ hour: 6, minute: 0, seconds: 0 });
    let times = [];

    do {
      time = time.plus({ hours: 1 })

      times.push({
        id: time.toMillis(),
        value: time.toFormat('HH:mm'),
        label: time.toFormat('ha').toLowerCase(),
      })
    } while(time.hour < 16)

    return times
  }

  return (
    <div className="flex justify-center bg-gray-100 h-full w-full">
      <div className="py-[58px] flex flex-col w-full max-w-[1344px] px-[15px] relative">
        <Alpha className="pr-[85px] md:pr-0">Available tee times</Alpha>

        <section className="flex flex-col w-full lg:flex-row mb-6 lg:mb-8 relative">
          <div className="flex flex-col w-full lg:w-[51%] lg:pr-8 mr-auto">
            <p className="mt-4 mb-[11px]">When would you like to play?</p>
            <div className="flex space-x-[18px] lg:space-x-8 mb-[15px] md:mb-0">
              <div className="flex flex-col w-[46%] lg:w-[192px]">
                <label className="uppercase text-[10px] sm:text-[13px] text-gray-500 font-bold tracking-widest">Date</label>
                <input
                  className="text-xs sm:text-base bg-transparent text-gray-500 border-0 border-b border-gray-500 border-opacity-20 pt-[6px] pb-[5px] pr-[4px] pl-[7px] min-h-[37px]"
                  min={getNextDay()}
                  type="date"
                  value={teeDate}
                  onChange={e => {
                      setTeeDate(e.target.value);
                      setLastUpdated(DateTime.now());
                    }
                  }
                />
              </div>

              <div className="flex flex-col w-[46%] lg:w-[192px]">
                <label className="uppercase text-[10px] sm:text-[13px] text-gray-500 font-bold tracking-widest">Tee Times From:</label>
                <CustomSelect
                  options={timeOptions()}
                  value={teeTime}
                  defaultValue={{ label: getNextTimeAsFormat('ha'), value: getNextTimeAsFormat('HH:mm') }}
                  onChange={option => {
                    setTeeTime({label: DateTime.fromFormat(teeDate + ' ' + option.value, 'yyyy-MM-dd HH:mm').toFormat('ha').toLowerCase(), value: option.value})
                    setLastUpdated(DateTime.now());
                  }}
                />
              </div>
            </div>
          </div>

          <div className="flex justify-bottom flex-shrink-0 lg:justify-end lg:flex-grow items-end mr-0">
            <button type="button" onClick={() => {
              setTeeDate(getNextDay());
              setTeeTime({ label: getNextTimeAsFormat('ha'), value: getNextTimeAsFormat('HH:mm') });

              setLastUpdated(DateTime.now());
            }} className="underline lg:mb-2.5 absolute right-0 -top-[18px] lg:static text-xs text-gray-500">Reset Filters</button>
          </div>
        </section>

        <section className="h-full">
          {loading && <div className="flex w-full h-full justify-center items-center">
          <svg xmlns="http://www.w3.org/2000/svg" xmlnsXlink="http://www.w3.org/1999/xlink" className="w-40 h-40" viewBox="0 0 100 100" preserveAspectRatio="xMidYMid">
            <circle className="opacity-20" cx="50" cy="50" r="41" stroke="#5C5C69" strokeWidth="1" fill="none"></circle>
            <circle cx="50" cy="50" r="41" stroke="#b4975a" strokeWidth="1" strokeLinecap="square" fill="none">
              <animateTransform attributeName="transform" type="rotate" repeatCount="indefinite" dur="1.2048192771084336s" values="0 50 50;180 50 50;720 50 50" keyTimes="0;0.5;1"></animateTransform>
              <animate attributeName="stroke-dasharray" repeatCount="indefinite" dur="1.2048192771084336s" values="25.761059759436304 231.8495378349267;128.8052987971815 128.8052987971815;25.761059759436304 231.8495378349267" keyTimes="0;0.5;1"></animate>
            </circle>
          </svg>
          </div>}
          {!loading && slots.length == 0 && <div className="text-gray-400 text-xl text-center">No slots available</div>}
          {!loading && <div className="grid md:grid-cols-3 lg:grid-cols-4 gap-2 md:gap-8 md:mt-8">
            {slots.map(slot => {
              let bookingDate = DateTime.fromISO(slot.date);

              return <div key={bookingDate.toMillis()} className="bg-white pt-[19px] pb-[23px] p-8 relative">
                <button
                  onClick={() => reserveSlot(bookingDate.toISO())}
                  className={`text-white absolute top-0 right-0 w-16 h-16 flex justify-center items-center bg-green-500`}
                >
                  <svg xmlns="http://www.w3.org/2000/svg" className="h-6 w-6" fill="none" viewBox="0 0 24 24" stroke="currentColor">
                    <path strokeLinecap="round" strokeLinejoin="round" strokeWidth="2" d="M14 5l7 7m0 0l-7 7m7-7H3" />
                  </svg>
                </button>
                <h3 className="text-lg">{bookingDate.toFormat('hh:mma')}</h3>

                <ul className="text-gray-600 mt-[34px] text-gray-500">
                  {slot.slots.map((slot, i) => {
                    return (<li key={`slot-${bookingDate.toMillis()}-${i}`} className="flex items-center space-x-2">
                      <svg xmlns="http://www.w3.org/2000/svg" className={`h-6 w-6 opacity-40 ${slot.name != 'Free' ? 'text-green-500' : ''}`} fill="none" viewBox="0 0 24 24">
                        <path fillRule="evenodd" clipRule="evenodd" d="M4.375 10H10.625C12.3509 10 13.75 11.3991 13.75 13.125C13.75 13.4702 13.4702 13.75 13.125 13.75C12.8045 13.75 12.5403 13.5087 12.5042 13.1979L12.4968 13.0148C12.4419 12.067 11.683 11.3081 10.7352 11.2532L10.625 11.25H4.375C3.33947 11.25 2.5 12.0895 2.5 13.125C2.5 13.4702 2.22018 13.75 1.875 13.75C1.52982 13.75 1.25 13.4702 1.25 13.125C1.25 11.4445 2.57644 10.0739 4.23944 10.0029L4.375 10H10.625H4.375ZM7.5 1.25C9.57107 1.25 11.25 2.92893 11.25 5C11.25 7.07107 9.57107 8.75 7.5 8.75C5.42893 8.75 3.75 7.07107 3.75 5C3.75 2.92893 5.42893 1.25 7.5 1.25ZM7.5 2.5C6.11929 2.5 5 3.61929 5 5C5 6.38071 6.11929 7.5 7.5 7.5C8.88071 7.5 10 6.38071 10 5C10 3.61929 8.88071 2.5 7.5 2.5Z" fill="currentColor"/>
                      </svg>

                      <p className="text-xs -mt-2">{slot.name}</p>
                    </li>)
                  })}
                </ul>
              </div>
            })}
          </div>}
        </section>
      </div>
    </div>
  )
}

Index.layout = page => <Main width='full' children={page} />

export default Index
